@@ -1,5 +1,5 @@
 /* Deal with interfaces.
-   Copyright (C) 2000, 2001, 2002, 2004, 2005, 2006, 2007
+   Copyright (C) 2000, 2001, 2002, 2004, 2005, 2006, 2007, 2008, 2009
    Free Software Foundation, Inc.
    Contributed by Andy Vaught
 
@@ -95,32 +95,32 @@ gfc_free_interface (gfc_interface *intr)
    minus respectively, leaving the rest unchanged.  */
 
 static gfc_intrinsic_op
-fold_unary (gfc_intrinsic_op operator)
+fold_unary (gfc_intrinsic_op op)
 {
-  switch (operator)
+  switch (op)
     {
     case INTRINSIC_UPLUS:
-      operator = INTRINSIC_PLUS;
+      op = INTRINSIC_PLUS;
       break;
     case INTRINSIC_UMINUS:
-      operator = INTRINSIC_MINUS;
+      op = INTRINSIC_MINUS;
       break;
     default:
       break;
     }
 
-  return operator;
+  return op;
 }
 
 
 /* Match a generic specification.  Depending on which type of
-   interface is found, the 'name' or 'operator' pointers may be set.
+   interface is found, the 'name' or 'op' pointers may be set.
    This subroutine doesn't return MATCH_NO.  */
 
 match
 gfc_match_generic_spec (interface_type *type,
 			char *name,
-			gfc_intrinsic_op *operator)
+			gfc_intrinsic_op *op)
 {
   char buffer[GFC_MAX_SYMBOL_LEN + 1];
   match m;
@@ -129,14 +129,14 @@ gfc_match_generic_spec (interface_type *type,
   if (gfc_match (" assignment ( = )") == MATCH_YES)
     {
       *type = INTERFACE_INTRINSIC_OP;
-      *operator = INTRINSIC_ASSIGN;
+      *op = INTRINSIC_ASSIGN;
       return MATCH_YES;
     }
 
   if (gfc_match (" operator ( %o )", &i) == MATCH_YES)
     {				/* Operator i/f */
       *type = INTERFACE_INTRINSIC_OP;
-      *operator = fold_unary (i);
+      *op = fold_unary (i);
       return MATCH_YES;
     }
 
@@ -184,12 +184,12 @@ gfc_match_interface (void)
   char name[GFC_MAX_SYMBOL_LEN + 1];
   interface_type type;
   gfc_symbol *sym;
-  gfc_intrinsic_op operator;
+  gfc_intrinsic_op op;
   match m;
 
   m = gfc_match_space ();
 
-  if (gfc_match_generic_spec (&type, name, &operator) == MATCH_ERROR)
+  if (gfc_match_generic_spec (&type, name, &op) == MATCH_ERROR)
     return MATCH_ERROR;
 
   /* If we're not looking at the end of the statement now, or if this
@@ -229,7 +229,7 @@ gfc_match_interface (void)
       break;
 
     case INTERFACE_INTRINSIC_OP:
-      current_interface.op = operator;
+      current_interface.op = op;
       break;
 
     case INTERFACE_NAMELESS:
@@ -275,12 +275,12 @@ gfc_match_end_interface (void)
 {
   char name[GFC_MAX_SYMBOL_LEN + 1];
   interface_type type;
-  gfc_intrinsic_op operator;
+  gfc_intrinsic_op op;
   match m;
 
   m = gfc_match_space ();
 
-  if (gfc_match_generic_spec (&type, name, &operator) == MATCH_ERROR)
+  if (gfc_match_generic_spec (&type, name, &op) == MATCH_ERROR)
     return MATCH_ERROR;
 
   /* If we're not looking at the end of the statement now, or if this
@@ -308,7 +308,7 @@ gfc_match_end_interface (void)
       break;
 
     case INTERFACE_INTRINSIC_OP:
-      if (type != current_interface.type || operator != current_interface.op)
+      if (type != current_interface.type || op != current_interface.op)
 	{
 
 	  if (current_interface.op == INTRINSIC_ASSIGN)
@@ -392,19 +392,19 @@ gfc_compare_derived_types (gfc_symbol *derived1, gfc_symbol *derived2)
       if (strcmp (dt1->name, dt2->name) != 0)
 	return 0;
 
-      if (dt1->access != dt2->access)
+      if (dt1->attr.access != dt2->attr.access)
 	return 0;
 
-      if (dt1->pointer != dt2->pointer)
+      if (dt1->attr.pointer != dt2->attr.pointer)
 	return 0;
 
-      if (dt1->dimension != dt2->dimension)
+      if (dt1->attr.dimension != dt2->attr.dimension)
 	return 0;
 
-     if (dt1->allocatable != dt2->allocatable)
+     if (dt1->attr.allocatable != dt2->attr.allocatable)
 	return 0;
 
-      if (dt1->dimension && gfc_compare_array_spec (dt1->as, dt2->as) == 0)
+      if (dt1->attr.dimension && gfc_compare_array_spec (dt1->as, dt2->as) == 0)
 	return 0;
 
       /* Make sure that link lists do not put this function into an 
@@ -479,7 +479,6 @@ compare_type_rank (gfc_symbol *s1, gfc_symbol *s2)
 }
 
 
-static int compare_interfaces (gfc_symbol *, gfc_symbol *, int);
 static int compare_intr_interfaces (gfc_symbol *, gfc_symbol *);
 
 /* Given two symbols that are formal arguments, compare their types
@@ -502,7 +501,7 @@ compare_type_rank_if (gfc_symbol *s1, gfc_symbol *s2)
     return 0;
 
   /* At this point, both symbols are procedures.  It can happen that
-     a external procedures are compared where one is identified by usage
+     external procedures are compared, where one is identified by usage
      to be a function or subroutine but the other is not.  Check TKR
      nonetheless for these cases.  */
   if (s1->attr.function == 0 && s1->attr.subroutine == 0)
@@ -547,7 +546,7 @@ find_keyword_arg (const char *name, gfc_formal_arglist *f)
    interfaces for that operator are legal.  */
 
 static void
-check_operator_interface (gfc_interface *intr, gfc_intrinsic_op operator)
+check_operator_interface (gfc_interface *intr, gfc_intrinsic_op op)
 {
   gfc_formal_arglist *formal;
   sym_intent i1, i2;
@@ -570,7 +569,7 @@ check_operator_interface (gfc_interface *intr, gfc_intrinsic_op operator)
       if (sym == NULL)
 	{
 	  gfc_error ("Alternate return cannot appear in operator "
-		     "interface at %L", &intr->where);
+		     "interface at %L", &intr->sym->declared_at);
 	  return;
 	}
       if (args == 0)
@@ -594,40 +593,45 @@ check_operator_interface (gfc_interface *intr, gfc_intrinsic_op operator)
 
   /* Only +, - and .not. can be unary operators.
      .not. cannot be a binary operator.  */
-  if (args == 0 || args > 2 || (args == 1 && operator != INTRINSIC_PLUS
-				&& operator != INTRINSIC_MINUS
-				&& operator != INTRINSIC_NOT)
-      || (args == 2 && operator == INTRINSIC_NOT))
+  if (args == 0 || args > 2 || (args == 1 && op != INTRINSIC_PLUS
+				&& op != INTRINSIC_MINUS
+				&& op != INTRINSIC_NOT)
+      || (args == 2 && op == INTRINSIC_NOT))
     {
       gfc_error ("Operator interface at %L has the wrong number of arguments",
-		 &intr->where);
+		 &intr->sym->declared_at);
       return;
     }
 
   /* Check that intrinsics are mapped to functions, except
      INTRINSIC_ASSIGN which should map to a subroutine.  */
-  if (operator == INTRINSIC_ASSIGN)
+  if (op == INTRINSIC_ASSIGN)
     {
       if (!sym->attr.subroutine)
 	{
 	  gfc_error ("Assignment operator interface at %L must be "
-		     "a SUBROUTINE", &intr->where);
+		     "a SUBROUTINE", &intr->sym->declared_at);
 	  return;
 	}
       if (args != 2)
 	{
 	  gfc_error ("Assignment operator interface at %L must have "
-		     "two arguments", &intr->where);
+		     "two arguments", &intr->sym->declared_at);
 	  return;
 	}
+
+      /* Allowed are (per F2003, 12.3.2.1.2 Defined assignments):
+         - First argument an array with different rank than second,
+         - Types and kinds do not conform, and
+         - First argument is of derived type.  */
       if (sym->formal->sym->ts.type != BT_DERIVED
-	  && sym->formal->next->sym->ts.type != BT_DERIVED
+	  && (r1 == 0 || r1 == r2)
 	  && (sym->formal->sym->ts.type == sym->formal->next->sym->ts.type
 	      || (gfc_numeric_ts (&sym->formal->sym->ts)
 		  && gfc_numeric_ts (&sym->formal->next->sym->ts))))
 	{
 	  gfc_error ("Assignment operator interface at %L must not redefine "
-		     "an INTRINSIC type assignment", &intr->where);
+		     "an INTRINSIC type assignment", &intr->sym->declared_at);
 	  return;
 	}
     }
@@ -636,31 +640,31 @@ check_operator_interface (gfc_interface *intr, gfc_intrinsic_op operator)
       if (!sym->attr.function)
 	{
 	  gfc_error ("Intrinsic operator interface at %L must be a FUNCTION",
-		     &intr->where);
+		     &intr->sym->declared_at);
 	  return;
 	}
     }
 
   /* Check intents on operator interfaces.  */
-  if (operator == INTRINSIC_ASSIGN)
+  if (op == INTRINSIC_ASSIGN)
     {
       if (i1 != INTENT_OUT && i1 != INTENT_INOUT)
 	gfc_error ("First argument of defined assignment at %L must be "
-		   "INTENT(IN) or INTENT(INOUT)", &intr->where);
+		   "INTENT(OUT) or INTENT(INOUT)", &intr->sym->declared_at);
 
       if (i2 != INTENT_IN)
 	gfc_error ("Second argument of defined assignment at %L must be "
-		   "INTENT(IN)", &intr->where);
+		   "INTENT(IN)", &intr->sym->declared_at);
     }
   else
     {
       if (i1 != INTENT_IN)
 	gfc_error ("First argument of operator interface at %L must be "
-		   "INTENT(IN)", &intr->where);
+		   "INTENT(IN)", &intr->sym->declared_at);
 
       if (args == 2 && i2 != INTENT_IN)
 	gfc_error ("Second argument of operator interface at %L must be "
-		   "INTENT(IN)", &intr->where);
+		   "INTENT(IN)", &intr->sym->declared_at);
     }
 
   /* From now on, all we have to do is check that the operator definition
@@ -678,7 +682,7 @@ check_operator_interface (gfc_interface *intr, gfc_intrinsic_op operator)
   ((t) == BT_INTEGER || (t) == BT_REAL || (t) == BT_COMPLEX)
 
   /* Unary ops are easy, do them first.  */
-  if (operator == INTRINSIC_NOT)
+  if (op == INTRINSIC_NOT)
     {
       if (t1 == BT_LOGICAL)
 	goto bad_repl;
@@ -686,7 +690,7 @@ check_operator_interface (gfc_interface *intr, gfc_intrinsic_op operator)
 	return;
     }
 
-  if (args == 1 && (operator == INTRINSIC_PLUS || operator == INTRINSIC_MINUS))
+  if (args == 1 && (op == INTRINSIC_PLUS || op == INTRINSIC_MINUS))
     {
       if (IS_NUMERIC_TYPE (t1))
 	goto bad_repl;
@@ -706,7 +710,7 @@ check_operator_interface (gfc_interface *intr, gfc_intrinsic_op operator)
   if (r1 != r2 && r1 != 0 && r2 != 0)
     return;
 
-  switch (operator)
+  switch (op)
   {
     case INTRINSIC_EQ:
     case INTRINSIC_EQ_OS:
@@ -799,7 +803,7 @@ count_types_test (gfc_formal_arglist *f1, gfc_formal_arglist *f2)
 
   /* Build an array of integers that gives the same integer to
      arguments of the same type/rank.  */
-  arg = gfc_getmem (n1 * sizeof (arginfo));
+  arg = XCNEWVEC (arginfo, n1);
 
   f = f1;
   for (i = 0; i < n1; i++, f = f->next)
@@ -958,8 +962,8 @@ generic_correspondence (gfc_formal_arglist *f1, gfc_formal_arglist *f2)
    We return nonzero if there exists an actual argument list that
    would be ambiguous between the two interfaces, zero otherwise.  */
 
-static int
-compare_interfaces (gfc_symbol *s1, gfc_symbol *s2, int generic_flag)
+int
+gfc_compare_interfaces (gfc_symbol *s1, gfc_symbol *s2, int generic_flag)
 {
   gfc_formal_arglist *f1, *f2;
 
@@ -1177,7 +1181,7 @@ check_interface1 (gfc_interface *p, gfc_interface *q0,
 	if (p->sym->name == q->sym->name && p->sym->module == q->sym->module)
 	  continue;
 
-	if (compare_interfaces (p->sym, q->sym, generic_flag))
+	if (gfc_compare_interfaces (p->sym, q->sym, generic_flag))
 	  {
 	    if (referenced)
 	      {
@@ -1247,7 +1251,7 @@ check_uop_interfaces (gfc_user_op *uop)
   gfc_namespace *ns;
 
   sprintf (interface_name, "operator interface '%s'", uop->name);
-  if (check_interface0 (uop->operator, interface_name))
+  if (check_interface0 (uop->op, interface_name))
     return;
 
   for (ns = gfc_current_ns; ns; ns = ns->parent)
@@ -1256,7 +1260,7 @@ check_uop_interfaces (gfc_user_op *uop)
       if (uop2 == NULL)
 	continue;
 
-      check_interface1 (uop->operator, uop2->operator, 0,
+      check_interface1 (uop->op, uop2->op, 0,
 			interface_name, true);
     }
 }
@@ -1292,76 +1296,76 @@ gfc_check_interfaces (gfc_namespace *ns)
 	sprintf (interface_name, "intrinsic '%s' operator",
 		 gfc_op2string (i));
 
-      if (check_interface0 (ns->operator[i], interface_name))
+      if (check_interface0 (ns->op[i], interface_name))
 	continue;
 
-      check_operator_interface (ns->operator[i], i);
+      check_operator_interface (ns->op[i], i);
 
       for (ns2 = ns; ns2; ns2 = ns2->parent)
 	{
-	  if (check_interface1 (ns->operator[i], ns2->operator[i], 0,
+	  if (check_interface1 (ns->op[i], ns2->op[i], 0,
 				interface_name, true))
 	    goto done;
 
 	  switch (i)
 	    {
 	      case INTRINSIC_EQ:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_EQ_OS],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_EQ_OS],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_EQ_OS:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_EQ],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_EQ],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_NE:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_NE_OS],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_NE_OS],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_NE_OS:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_NE],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_NE],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_GT:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_GT_OS],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_GT_OS],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_GT_OS:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_GT],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_GT],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_GE:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_GE_OS],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_GE_OS],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_GE_OS:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_GE],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_GE],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_LT:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_LT_OS],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_LT_OS],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_LT_OS:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_LT],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_LT],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_LE:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_LE_OS],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_LE_OS],
 				      0, interface_name, true)) goto done;
 		break;
 
 	      case INTRINSIC_LE_OS:
-		if (check_interface1 (ns->operator[i], ns2->operator[INTRINSIC_LE],
+		if (check_interface1 (ns->op[i], ns2->op[INTRINSIC_LE],
 				      0, interface_name, true)) goto done;
 		break;
 
@@ -1464,7 +1468,7 @@ compare_parameter (gfc_symbol *formal, gfc_expr *actual,
 	 if (!compare_intr_interfaces (formal, actual->symtree->n.sym))
 	   goto proc_fail;
 	}
-      else if (!compare_interfaces (formal, actual->symtree->n.sym, 0))
+      else if (!gfc_compare_interfaces (formal, actual->symtree->n.sym, 0))
 	goto proc_fail;
 
       return 1;
@@ -1569,7 +1573,7 @@ compare_parameter_protected (gfc_symbol *formal, gfc_expr *actual)
   if (actual->expr_type != EXPR_VARIABLE)
     return 1;
 
-  if (!actual->symtree->n.sym->attr.protected)
+  if (!actual->symtree->n.sym->attr.is_protected)
     return 1;
 
   if (!actual->symtree->n.sym->attr.use_assoc)
@@ -1825,9 +1829,9 @@ has_vector_subscript (gfc_expr *e)
 
 static int
 compare_actual_formal (gfc_actual_arglist **ap, gfc_formal_arglist *formal,
-		       int ranks_must_agree, int is_elemental, locus *where)
+	 	       int ranks_must_agree, int is_elemental, locus *where)
 {
-  gfc_actual_arglist **new, *a, *actual, temp;
+  gfc_actual_arglist **new_arg, *a, *actual, temp;
   gfc_formal_arglist *f;
   int i, n, na;
   unsigned long actual_size, formal_size;
@@ -1841,10 +1845,10 @@ compare_actual_formal (gfc_actual_arglist **ap, gfc_formal_arglist *formal,
   for (f = formal; f; f = f->next)
     n++;
 
-  new = (gfc_actual_arglist **) alloca (n * sizeof (gfc_actual_arglist *));
+  new_arg = (gfc_actual_arglist **) alloca (n * sizeof (gfc_actual_arglist *));
 
   for (i = 0; i < n; i++)
-    new[i] = NULL;
+    new_arg[i] = NULL;
 
   na = 0;
   f = formal;
@@ -1872,7 +1876,7 @@ compare_actual_formal (gfc_actual_arglist **ap, gfc_formal_arglist *formal,
 	      return 0;
 	    }
 
-	  if (new[i] != NULL)
+	  if (new_arg[i] != NULL)
 	    {
 	      if (where)
 		gfc_error ("Keyword argument '%s' at %L is already associated "
@@ -1963,6 +1967,17 @@ compare_actual_formal (gfc_actual_arglist **ap, gfc_formal_arglist *formal,
 	  return  0;
 	}
 
+      /* Satisfy 12.4.1.3 by ensuring that a procedure pointer actual argument
+	 is provided for a procedure pointer formal argument.  */
+      if (f->sym->attr.proc_pointer
+	  && !a->expr->symtree->n.sym->attr.proc_pointer)
+	{
+	  if (where)
+	    gfc_error ("Expected a procedure pointer for argument '%s' at %L",
+		       f->sym->name, &a->expr->where);
+	  return 0;
+	}
+
       /* Satisfy 12.4.1.2 by ensuring that a procedure actual argument is
 	 provided for a procedure formal argument.  */
       if (a->expr->ts.type != BT_PROCEDURE
@@ -2106,14 +2121,14 @@ compare_actual_formal (gfc_actual_arglist **ap, gfc_formal_arglist *formal,
       if (a == actual)
 	na = i;
 
-      new[i++] = a;
+      new_arg[i++] = a;
     }
 
   /* Make sure missing actual arguments are optional.  */
   i = 0;
   for (f = formal; f; f = f->next, i++)
     {
-      if (new[i] != NULL)
+      if (new_arg[i] != NULL)
 	continue;
       if (f->sym == NULL)
 	{
@@ -2135,27 +2150,27 @@ compare_actual_formal (gfc_actual_arglist **ap, gfc_formal_arglist *formal,
      argument list with null arguments in the right places.  The head
      of the list remains the head.  */
   for (i = 0; i < n; i++)
-    if (new[i] == NULL)
-      new[i] = gfc_get_actual_arglist ();
+    if (new_arg[i] == NULL)
+      new_arg[i] = gfc_get_actual_arglist ();
 
   if (na != 0)
     {
-      temp = *new[0];
-      *new[0] = *actual;
+      temp = *new_arg[0];
+      *new_arg[0] = *actual;
       *actual = temp;
 
-      a = new[0];
-      new[0] = new[na];
-      new[na] = a;
+      a = new_arg[0];
+      new_arg[0] = new_arg[na];
+      new_arg[na] = a;
     }
 
   for (i = 0; i < n - 1; i++)
-    new[i]->next = new[i + 1];
+    new_arg[i]->next = new_arg[i + 1];
 
-  new[i]->next = NULL;
+  new_arg[i]->next = NULL;
 
   if (*ap == NULL && n > 0)
-    *ap = new[0];
+    *ap = new_arg[0];
 
   /* Note the types of omitted optional arguments.  */
   for (a = *ap, f = formal; a; a = a->next, f = f->next)
@@ -2211,7 +2226,7 @@ pair_cmp (const void *p1, const void *p2)
    refer to the same expression. The analysis is conservative.
    Returning FAILURE will produce no warning.  */
 
-static try
+static gfc_try
 compare_actual_expr (gfc_expr *e1, gfc_expr *e2)
 {
   const gfc_ref *r1, *r2;
@@ -2260,7 +2275,7 @@ compare_actual_expr (gfc_expr *e1, gfc_expr *e2)
    another, check that identical actual arguments aren't not
    associated with some incompatible INTENTs.  */
 
-static try
+static gfc_try
 check_some_aliasing (gfc_formal_arglist *f, gfc_actual_arglist *a)
 {
   sym_intent f1_intent, f2_intent;
@@ -2268,7 +2283,7 @@ check_some_aliasing (gfc_formal_arglist *f, gfc_actual_arglist *a)
   gfc_actual_arglist *a1;
   size_t n, i, j;
   argpair *p;
-  try t = SUCCESS;
+  gfc_try t = SUCCESS;
 
   n = 0;
   for (f1 = f, a1 = a;; f1 = f1->next, a1 = a1->next)
@@ -2348,7 +2363,7 @@ compare_parameter_intent (gfc_symbol *formal, gfc_expr *actual)
    another, check that they are compatible in the sense that intents
    are not mismatched.  */
 
-static try
+static gfc_try
 check_intents (gfc_formal_arglist *f, gfc_actual_arglist *a)
 {
   sym_intent f_intent;
@@ -2405,19 +2420,22 @@ void
 gfc_procedure_use (gfc_symbol *sym, gfc_actual_arglist **ap, locus *where)
 {
 
-  /* Warn about calls with an implicit interface.  */
+  /* Warn about calls with an implicit interface.  Special case
+     for calling a ISO_C_BINDING becase c_loc and c_funloc
+     are pseudo-unknown.  */
   if (gfc_option.warn_implicit_interface
-      && sym->attr.if_source == IFSRC_UNKNOWN)
+      && sym->attr.if_source == IFSRC_UNKNOWN
+      && ! sym->attr.is_iso_c)
     gfc_warning ("Procedure '%s' called with an implicit interface at %L",
 		 sym->name, where);
 
-  if (sym->interface && sym->interface->attr.intrinsic)
+  if (sym->ts.interface && sym->ts.interface->attr.intrinsic)
     {
       gfc_intrinsic_sym *isym;
-      isym = gfc_find_function (sym->interface->name);
+      isym = gfc_find_function (sym->ts.interface->name);
       if (isym != NULL)
 	{
-	  if (compare_actual_formal_intr (ap, sym->interface))
+	  if (compare_actual_formal_intr (ap, sym->ts.interface))
 	    return;
 	  gfc_error ("Type/rank mismatch in argument '%s' at %L",
 		     sym->name, where);
@@ -2425,9 +2443,24 @@ gfc_procedure_use (gfc_symbol *sym, gfc_actual_arglist **ap, locus *where)
 	}
     }
 
-  if (sym->attr.if_source == IFSRC_UNKNOWN
-      || !compare_actual_formal (ap, sym->formal, 0,
-				 sym->attr.elemental, where))
+  if (sym->attr.if_source == IFSRC_UNKNOWN)
+    {
+      gfc_actual_arglist *a;
+      for (a = *ap; a; a = a->next)
+	{
+	  /* Skip g77 keyword extensions like %VAL, %REF, %LOC.  */
+	  if (a->name != NULL && a->name[0] != '%')
+	    {
+	      gfc_error("Keyword argument requires explicit interface "
+			"for procedure '%s' at %L", sym->name, &a->expr->where);
+	      break;
+	    }
+	}
+
+      return;
+    }
+
+  if (!compare_actual_formal (ap, sym->formal, 0, sym->attr.elemental, where))
     return;
 
   check_intents (sym->formal, *ap);
@@ -2436,6 +2469,30 @@ gfc_procedure_use (gfc_symbol *sym, gfc_actual_arglist **ap, locus *where)
 }
 
 
+/* Try if an actual argument list matches the formal list of a symbol,
+   respecting the symbol's attributes like ELEMENTAL.  This is used for
+   GENERIC resolution.  */
+
+bool
+gfc_arglist_matches_symbol (gfc_actual_arglist** args, gfc_symbol* sym)
+{
+  bool r;
+
+  gcc_assert (sym->attr.flavor == FL_PROCEDURE);
+
+  r = !sym->attr.elemental;
+  if (compare_actual_formal (args, sym->formal, r, !r, NULL))
+    {
+      check_intents (sym->formal, *args);
+      if (gfc_option.warn_aliasing)
+	check_some_aliasing (sym->formal, *args);
+      return true;
+    }
+
+  return false;
+}
+
+
 /* Given an interface pointer and an actual argument list, search for
    a formal argument list that matches the actual.  If found, returns
    a pointer to the symbol of the correct interface.  Returns NULL if
@@ -2445,8 +2502,6 @@ gfc_symbol *
 gfc_search_interface (gfc_interface *intr, int sub_flag,
 		      gfc_actual_arglist **ap)
 {
-  int r;
-
   for (; intr; intr = intr->next)
     {
       if (sub_flag && intr->sym->attr.function)
@@ -2454,15 +2509,8 @@ gfc_search_interface (gfc_interface *intr, int sub_flag,
       if (!sub_flag && intr->sym->attr.subroutine)
 	continue;
 
-      r = !intr->sym->attr.elemental;
-
-      if (compare_actual_formal (ap, intr->sym->formal, r, !r, NULL))
-	{
-	  check_intents (intr->sym->formal, *ap);
-	  if (gfc_option.warn_aliasing)
-	    check_some_aliasing (intr->sym->formal, *ap);
-	  return intr->sym;
-	}
+      if (gfc_arglist_matches_symbol (ap, intr->sym))
+	return intr->sym;
     }
 
   return NULL;
@@ -2490,8 +2538,8 @@ find_symtree0 (gfc_symtree *root, gfc_symbol *sym)
 
 /* Find a symtree for a symbol.  */
 
-static gfc_symtree *
-find_sym_in_symtree (gfc_symbol *sym)
+gfc_symtree *
+gfc_find_sym_in_symtree (gfc_symbol *sym)
 {
   gfc_symtree *st;
   gfc_namespace *ns;
@@ -2523,7 +2571,7 @@ find_sym_in_symtree (gfc_symbol *sym)
    interface.  If one is found, the expression node is replaced with
    the appropriate function call.  */
 
-try
+gfc_try
 gfc_extend_expr (gfc_expr *e)
 {
   gfc_actual_arglist *actual;
@@ -2543,7 +2591,7 @@ gfc_extend_expr (gfc_expr *e)
       actual->next->expr = e->value.op.op2;
     }
 
-  i = fold_unary (e->value.op.operator);
+  i = fold_unary (e->value.op.op);
 
   if (i == INTRINSIC_USER)
     {
@@ -2553,7 +2601,7 @@ gfc_extend_expr (gfc_expr *e)
 	  if (uop == NULL)
 	    continue;
 
-	  sym = gfc_search_interface (uop->operator, 0, &actual);
+	  sym = gfc_search_interface (uop->op, 0, &actual);
 	  if (sym != NULL)
 	    break;
 	}
@@ -2568,48 +2616,48 @@ gfc_extend_expr (gfc_expr *e)
 	    {
 	      case INTRINSIC_EQ:
 	      case INTRINSIC_EQ_OS:
-		sym = gfc_search_interface (ns->operator[INTRINSIC_EQ], 0, &actual);
+		sym = gfc_search_interface (ns->op[INTRINSIC_EQ], 0, &actual);
 		if (sym == NULL)
-		  sym = gfc_search_interface (ns->operator[INTRINSIC_EQ_OS], 0, &actual);
+		  sym = gfc_search_interface (ns->op[INTRINSIC_EQ_OS], 0, &actual);
 		break;
 
 	      case INTRINSIC_NE:
 	      case INTRINSIC_NE_OS:
-		sym = gfc_search_interface (ns->operator[INTRINSIC_NE], 0, &actual);
+		sym = gfc_search_interface (ns->op[INTRINSIC_NE], 0, &actual);
 		if (sym == NULL)
-		  sym = gfc_search_interface (ns->operator[INTRINSIC_NE_OS], 0, &actual);
+		  sym = gfc_search_interface (ns->op[INTRINSIC_NE_OS], 0, &actual);
 		break;
 
 	      case INTRINSIC_GT:
 	      case INTRINSIC_GT_OS:
-		sym = gfc_search_interface (ns->operator[INTRINSIC_GT], 0, &actual);
+		sym = gfc_search_interface (ns->op[INTRINSIC_GT], 0, &actual);
 		if (sym == NULL)
-		  sym = gfc_search_interface (ns->operator[INTRINSIC_GT_OS], 0, &actual);
+		  sym = gfc_search_interface (ns->op[INTRINSIC_GT_OS], 0, &actual);
 		break;
 
 	      case INTRINSIC_GE:
 	      case INTRINSIC_GE_OS:
-		sym = gfc_search_interface (ns->operator[INTRINSIC_GE], 0, &actual);
+		sym = gfc_search_interface (ns->op[INTRINSIC_GE], 0, &actual);
 		if (sym == NULL)
-		  sym = gfc_search_interface (ns->operator[INTRINSIC_GE_OS], 0, &actual);
+		  sym = gfc_search_interface (ns->op[INTRINSIC_GE_OS], 0, &actual);
 		break;
 
 	      case INTRINSIC_LT:
 	      case INTRINSIC_LT_OS:
-		sym = gfc_search_interface (ns->operator[INTRINSIC_LT], 0, &actual);
+		sym = gfc_search_interface (ns->op[INTRINSIC_LT], 0, &actual);
 		if (sym == NULL)
-		  sym = gfc_search_interface (ns->operator[INTRINSIC_LT_OS], 0, &actual);
+		  sym = gfc_search_interface (ns->op[INTRINSIC_LT_OS], 0, &actual);
 		break;
 
 	      case INTRINSIC_LE:
 	      case INTRINSIC_LE_OS:
-		sym = gfc_search_interface (ns->operator[INTRINSIC_LE], 0, &actual);
+		sym = gfc_search_interface (ns->op[INTRINSIC_LE], 0, &actual);
 		if (sym == NULL)
-		  sym = gfc_search_interface (ns->operator[INTRINSIC_LE_OS], 0, &actual);
+		  sym = gfc_search_interface (ns->op[INTRINSIC_LE_OS], 0, &actual);
 		break;
 
 	      default:
-		sym = gfc_search_interface (ns->operator[i], 0, &actual);
+		sym = gfc_search_interface (ns->op[i], 0, &actual);
 	    }
 
 	  if (sym != NULL)
@@ -2629,7 +2677,7 @@ gfc_extend_expr (gfc_expr *e)
 
   /* Change the expression node to a function call.  */
   e->expr_type = EXPR_FUNCTION;
-  e->symtree = find_sym_in_symtree (sym);
+  e->symtree = gfc_find_sym_in_symtree (sym);
   e->value.function.actual = actual;
   e->value.function.esym = NULL;
   e->value.function.isym = NULL;
@@ -2655,7 +2703,7 @@ gfc_extend_expr (gfc_expr *e)
    SUCCESS if the node was replaced.  On FAILURE, no error is
    generated.  */
 
-try
+gfc_try
 gfc_extend_assign (gfc_code *c, gfc_namespace *ns)
 {
   gfc_actual_arglist *actual;
@@ -2666,7 +2714,8 @@ gfc_extend_assign (gfc_code *c, gfc_namespace *ns)
   rhs = c->expr2;
 
   /* Don't allow an intrinsic assignment to be replaced.  */
-  if (lhs->ts.type != BT_DERIVED && rhs->ts.type != BT_DERIVED
+  if (lhs->ts.type != BT_DERIVED
+      && (rhs->rank == 0 || rhs->rank == lhs->rank)
       && (lhs->ts.type == rhs->ts.type
 	  || (gfc_numeric_ts (&lhs->ts) && gfc_numeric_ts (&rhs->ts))))
     return FAILURE;
@@ -2681,7 +2730,7 @@ gfc_extend_assign (gfc_code *c, gfc_namespace *ns)
 
   for (; ns; ns = ns->parent)
     {
-      sym = gfc_search_interface (ns->operator[INTRINSIC_ASSIGN], 1, &actual);
+      sym = gfc_search_interface (ns->op[INTRINSIC_ASSIGN], 1, &actual);
       if (sym != NULL)
 	break;
     }
@@ -2695,7 +2744,7 @@ gfc_extend_assign (gfc_code *c, gfc_namespace *ns)
 
   /* Replace the assignment with the call.  */
   c->op = EXEC_ASSIGN_CALL;
-  c->symtree = find_sym_in_symtree (sym);
+  c->symtree = gfc_find_sym_in_symtree (sym);
   c->expr = NULL;
   c->expr2 = NULL;
   c->ext.actual = actual;
@@ -2708,17 +2757,17 @@ gfc_extend_assign (gfc_code *c, gfc_namespace *ns)
    the given interface list.  Ambiguity isn't checked yet since module
    procedures can be present without interfaces.  */
 
-static try
-check_new_interface (gfc_interface *base, gfc_symbol *new)
+static gfc_try
+check_new_interface (gfc_interface *base, gfc_symbol *new_sym)
 {
   gfc_interface *ip;
 
   for (ip = base; ip; ip = ip->next)
     {
-      if (ip->sym == new)
+      if (ip->sym == new_sym)
 	{
 	  gfc_error ("Entity '%s' at %C is already present in the interface",
-		     new->name);
+		     new_sym->name);
 	  return FAILURE;
 	}
     }
@@ -2729,8 +2778,8 @@ check_new_interface (gfc_interface *base, gfc_symbol *new)
 
 /* Add a symbol to the current interface.  */
 
-try
-gfc_add_interface (gfc_symbol *new)
+gfc_try
+gfc_add_interface (gfc_symbol *new_sym)
 {
   gfc_interface **head, *intr;
   gfc_namespace *ns;
@@ -2748,52 +2797,52 @@ gfc_add_interface (gfc_symbol *new)
 	  {
 	    case INTRINSIC_EQ:
 	    case INTRINSIC_EQ_OS:
-	      if (check_new_interface (ns->operator[INTRINSIC_EQ], new) == FAILURE ||
-	          check_new_interface (ns->operator[INTRINSIC_EQ_OS], new) == FAILURE)
+	      if (check_new_interface (ns->op[INTRINSIC_EQ], new_sym) == FAILURE ||
+	          check_new_interface (ns->op[INTRINSIC_EQ_OS], new_sym) == FAILURE)
 		return FAILURE;
 	      break;
 
 	    case INTRINSIC_NE:
 	    case INTRINSIC_NE_OS:
-	      if (check_new_interface (ns->operator[INTRINSIC_NE], new) == FAILURE ||
-	          check_new_interface (ns->operator[INTRINSIC_NE_OS], new) == FAILURE)
+	      if (check_new_interface (ns->op[INTRINSIC_NE], new_sym) == FAILURE ||
+	          check_new_interface (ns->op[INTRINSIC_NE_OS], new_sym) == FAILURE)
 		return FAILURE;
 	      break;
 
 	    case INTRINSIC_GT:
 	    case INTRINSIC_GT_OS:
-	      if (check_new_interface (ns->operator[INTRINSIC_GT], new) == FAILURE ||
-	          check_new_interface (ns->operator[INTRINSIC_GT_OS], new) == FAILURE)
+	      if (check_new_interface (ns->op[INTRINSIC_GT], new_sym) == FAILURE ||
+	          check_new_interface (ns->op[INTRINSIC_GT_OS], new_sym) == FAILURE)
 		return FAILURE;
 	      break;
 
 	    case INTRINSIC_GE:
 	    case INTRINSIC_GE_OS:
-	      if (check_new_interface (ns->operator[INTRINSIC_GE], new) == FAILURE ||
-	          check_new_interface (ns->operator[INTRINSIC_GE_OS], new) == FAILURE)
+	      if (check_new_interface (ns->op[INTRINSIC_GE], new_sym) == FAILURE ||
+	          check_new_interface (ns->op[INTRINSIC_GE_OS], new_sym) == FAILURE)
 		return FAILURE;
 	      break;
 
 	    case INTRINSIC_LT:
 	    case INTRINSIC_LT_OS:
-	      if (check_new_interface (ns->operator[INTRINSIC_LT], new) == FAILURE ||
-	          check_new_interface (ns->operator[INTRINSIC_LT_OS], new) == FAILURE)
+	      if (check_new_interface (ns->op[INTRINSIC_LT], new_sym) == FAILURE ||
+	          check_new_interface (ns->op[INTRINSIC_LT_OS], new_sym) == FAILURE)
 		return FAILURE;
 	      break;
 
 	    case INTRINSIC_LE:
 	    case INTRINSIC_LE_OS:
-	      if (check_new_interface (ns->operator[INTRINSIC_LE], new) == FAILURE ||
-	          check_new_interface (ns->operator[INTRINSIC_LE_OS], new) == FAILURE)
+	      if (check_new_interface (ns->op[INTRINSIC_LE], new_sym) == FAILURE ||
+	          check_new_interface (ns->op[INTRINSIC_LE_OS], new_sym) == FAILURE)
 		return FAILURE;
 	      break;
 
 	    default:
-	      if (check_new_interface (ns->operator[current_interface.op], new) == FAILURE)
+	      if (check_new_interface (ns->op[current_interface.op], new_sym) == FAILURE)
 		return FAILURE;
 	  }
 
-      head = &current_interface.ns->operator[current_interface.op];
+      head = &current_interface.ns->op[current_interface.op];
       break;
 
     case INTERFACE_GENERIC:
@@ -2803,7 +2852,7 @@ gfc_add_interface (gfc_symbol *new)
 	  if (sym == NULL)
 	    continue;
 
-	  if (check_new_interface (sym->generic, new) == FAILURE)
+	  if (check_new_interface (sym->generic, new_sym) == FAILURE)
 	    return FAILURE;
 	}
 
@@ -2811,11 +2860,11 @@ gfc_add_interface (gfc_symbol *new)
       break;
 
     case INTERFACE_USER_OP:
-      if (check_new_interface (current_interface.uop->operator, new)
+      if (check_new_interface (current_interface.uop->op, new_sym)
 	  == FAILURE)
 	return FAILURE;
 
-      head = &current_interface.uop->operator;
+      head = &current_interface.uop->op;
       break;
 
     default:
@@ -2823,7 +2872,7 @@ gfc_add_interface (gfc_symbol *new)
     }
 
   intr = gfc_get_interface ();
-  intr->sym = new;
+  intr->sym = new_sym;
   intr->where = gfc_current_locus;
 
   intr->next = *head;
@@ -2839,7 +2888,7 @@ gfc_current_interface_head (void)
   switch (current_interface.type)
     {
       case INTERFACE_INTRINSIC_OP:
-	return current_interface.ns->operator[current_interface.op];
+	return current_interface.ns->op[current_interface.op];
 	break;
 
       case INTERFACE_GENERIC:
@@ -2847,7 +2896,7 @@ gfc_current_interface_head (void)
 	break;
 
       case INTERFACE_USER_OP:
-	return current_interface.uop->operator;
+	return current_interface.uop->op;
 	break;
 
       default:
@@ -2862,7 +2911,7 @@ gfc_set_current_interface_head (gfc_interface *i)
   switch (current_interface.type)
     {
       case INTERFACE_INTRINSIC_OP:
-	current_interface.ns->operator[current_interface.op] = i;
+	current_interface.ns->op[current_interface.op] = i;
 	break;
 
       case INTERFACE_GENERIC:
@@ -2870,7 +2919,7 @@ gfc_set_current_interface_head (gfc_interface *i)
 	break;
 
       case INTERFACE_USER_OP:
-	current_interface.uop->operator = i;
+	current_interface.uop->op = i;
 	break;
 
       default:
